#!/usr/bin/env python3

import logging
import array
from enum import Enum
import dbus
import dbus.exceptions
import dbus.mainloop.glib
import dbus.service

from ble import (
    Advertisement,
    Characteristic,
    Service,
    Application,
    find_adapter,
    Descriptor,
    Agent,
)

from wifi import (
    WifiConnectionManager,
)

from device import (
    DeviceManager,
)

BLUEZ_SERVICE_NAME = "org.bluez"
GATT_MANAGER_IFACE = "org.bluez.GattManager1"
GATT_CHRC_IFACE =    'org.bluez.GattCharacteristic1'
LE_ADVERTISEMENT_IFACE = "org.bluez.LEAdvertisement1"
LE_ADVERTISING_MANAGER_IFACE = "org.bluez.LEAdvertisingManager1"

MainLoop = None
try:
    from gi.repository import GLib

    MainLoop = GLib.MainLoop
except ImportError:
    import gobject as GObject

    MainLoop = GObject.MainLoop

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logHandler = logging.StreamHandler()
filelogHandler = logging.FileHandler("logs.log")
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
logHandler.setFormatter(formatter)
filelogHandler.setFormatter(formatter)
logger.addHandler(filelogHandler)
logger.addHandler(logHandler)

wifiManager = WifiConnectionManager()

deviceManager = DeviceManager()

def string_to_dbus_array(input_string):
    # Convert the string to a list of integers (ASCII values)
    byte_list = [dbus.Byte(ord(char)) for char in input_string]

    # Create a D-Bus array from the list
    dbus_array = dbus.Array(byte_list, signature=dbus.Signature('y'))

    return dbus_array

def dbus_array_to_string(dbus_array):
    # Convert the dbus array to bytes
    byte_array = bytes(int(byte) for byte in dbus_array)
    # Decode the bytes as UTF-8
    return byte_array.decode('utf-8')

def register_app_cb():
    logger.info("GATT application registered")


def register_app_error_cb(error):
    logger.critical("Failed to register application: %s", str(error))
    mainloop.quit()

class TeddyS1Service(Service):
    """
    Dummy test service that provides characteristics and descriptors that
    exercise various API functionality.

    """

    TEDDY_SVC_UUID = "0d42b224-7dff-4558-9cc2-ac7c90ae00ba"

    def __init__(self, bus, index):
        Service.__init__(self, bus, index, self.TEDDY_SVC_UUID, True)
        self.add_characteristic(WifiSsidCharacteristic(bus, 0, self))
        self.add_characteristic(WifiPskCharacteristic(bus, 1, self))
        self.wifi_status_characteristic = WifiConnectionStatusCharacteristic(bus, 2, self)
        self.add_characteristic(self.wifi_status_characteristic)
        self.add_characteristic(DeviceIdCharacteristic(bus, 3, self))
        self.add_characteristic(TokenCharacteristic(bus, 4, self))
    
    def UpdateWifiStatus(self, new_state):
        if isinstance(new_state, WifiConnectionStatusCharacteristic.State):
            self.wifi_status_characteristic.current_state = new_state
            self.wifi_status_characteristic.value = new_state.to_bytearray()
            self.wifi_status_characteristic.notify_connection_status()
        else:
            logger.error("Invalid WiFi connection state")

class WifiSsidCharacteristic(Characteristic):
    uuid = "96b111e2-8446-4636-bc57-841ffbd675ab"
    description = b"Get/set wifi ssid"

    def __init__(self, bus, index, service):
        Characteristic.__init__(
            self, bus, index, self.uuid, ["read", "write"], service,
        )
    
        self.value = [0xFF]
        self.add_descriptor(CharacteristicUserDescriptionDescriptor(bus, 1, self))
    
    def ReadValue(self, options):
        logger.debug("SSID Read: %s", dbus_array_to_string(self.value))
        return self.value

    def WriteValue(self, value, options):
        ssidStr = dbus_array_to_string(value)
        logger.debug(f"SSID Write: {ssidStr}")

        wifiManager.ssid = ssidStr

        self.value = value

class WifiPskCharacteristic(Characteristic):
    uuid = "3d77b718-103d-430e-8212-a75cce5e199a"
    description = b"Set wifi psk"

    def __init__(self, bus, index, service):
        Characteristic.__init__(
            self, bus, index, self.uuid, ["write"], service,
        )

        self.value = "password"
        self.add_descriptor(CharacteristicUserDescriptionDescriptor(bus, 1, self))

    def WriteValue(self, value, options):
        pskStr = dbus_array_to_string(value)
        logger.debug("PSK Write: %s", pskStr)

        # Attempt to connect to wifi network
        wifiManager.psk = pskStr
        connected = wifiManager.connect_to_wifi()

        # Send notification whether connection was successful/failed
        if connected:
            self.service.UpdateWifiStatus(WifiConnectionStatusCharacteristic.State.CONNECTED)
            # Check for firmware update
            wifiManager.update_and_upgrade()
        else:
            self.service.UpdateWifiStatus(WifiConnectionStatusCharacteristic.State.CONNECTION_FAILED)

        self.value = value

class WifiConnectionStatusCharacteristic(Characteristic):
    uuid = "179d7379-14ad-427d-b273-7db7ce854f15"
    description = b"Get wifi connection status {'NOT_CONFIGURED', 'CONNECTING', 'CONNECTED', 'CONNECTION_FAILED'}"

    class State(Enum):
        NOT_CONFIGURED = 0
        CONNECTING = 1
        CONNECTED = 2
        CONNECTION_FAILED = 3

        def to_bytearray(self):
            return bytearray([self.value])

        @classmethod
        def has_value(cls, value):
            return value in cls._value2member_map_    

    def __init__(self, bus, index, service):
        Characteristic.__init__(
            self, bus, index, self.uuid, ["read", "notify"], service,
        )
        self.notifying = False  # is a phone connected and subscribed to this characteristic
        self.current_state = self.State.NOT_CONFIGURED
        self.value = self.current_state.to_bytearray()
        self.add_descriptor(CharacteristicUserDescriptionDescriptor(bus, 1, self))

    def ReadValue(self, options):
        logger.debug("Wifi Status Read: %s", repr(self.value))

        return self.current_state.to_bytearray()

    def notify_connection_status(self):
        if not self.notifying:
            return

        dbus_value = [dbus.Byte(b) for b in self.value]
        self.PropertiesChanged(
                GATT_CHRC_IFACE,
                { 'Value': dbus_value }, [])

    def StartNotify(self):
        if self.notifying:
            logger.debug('Already notifying, nothing to do')
            return

        self.notifying = True
        self.notify_connection_status()

    def StopNotify(self):
        if not self.notifying:
            logger.debug('Not notifying, nothing to do')
            return

        self.notifying = False

class DeviceIdCharacteristic(Characteristic):
    uuid = "d4b78e5f-455b-4b3b-8d71-e8e50f487901"
    description = b"Get Device ID"

    def __init__(self, bus, index, service):
        Characteristic.__init__(
            self, bus, index, self.uuid, ["read"], service,
        )

        self.value = [0xFF]
        self.add_descriptor(CharacteristicUserDescriptionDescriptor(bus, 1, self))

    def ReadValue(self, options):
        self.value = string_to_dbus_array(deviceManager.deviceId)
        logger.debug("Device ID Read: %s", deviceManager.deviceId)
        return self.value

class TokenCharacteristic(Characteristic):
    uuid = "f3b50b22-29f4-4abe-a969-e9f2b755a891"
    description = b"Write the device JWT needed for authenticated requests to the server"

    class ValidationState:
        SUCCESS = "TOKEN_VALID"
        INVALID = "TOKEN_INVALID"
        SAVE_ERROR = "SAVE_ERROR"
        NETWORK_ERROR = "NETWORK_ERROR"

    def __init__(self, bus, index, service):
        Characteristic.__init__(
            self, bus, index, self.uuid, ["write", "notify"], service,
        )
        self.notifying = False
        self.add_descriptor(CharacteristicUserDescriptionDescriptor(bus, 1, self))

    def WriteValue(self, value, options):
        token = dbus_array_to_string(value)
        logger.debug("Token Write: %s", token)
        
        # Try to save the token first
        try:
            deviceManager.save_device_token(token)
        except Exception as e:
            logger.error(f"Failed to save token: {e}")
            self.notify_validation_result(self.ValidationState.SAVE_ERROR)
            return

        # Validate token with server
        try:
            if deviceManager.validate_device_token(token):
                self.notify_validation_result(self.ValidationState.SUCCESS)
                logger.debug("Token validation successful")
            else:
                self.notify_validation_result(self.ValidationState.INVALID)
                logger.warning("Token validation failed")
        except Exception as e:
            logger.error(f"Network error during validation: {e}")
            self.notify_validation_result(self.ValidationState.NETWORK_ERROR)

    def notify_validation_result(self, result):
        if not self.notifying:
            logger.debug("Not notifying - client hasn't subscribed")
            return

        logger.debug(f"Notifying validation result: {result}")
        dbus_value = string_to_dbus_array(result)
        self.PropertiesChanged(
                GATT_CHRC_IFACE,
                { 'Value': dbus_value }, [])

    def StartNotify(self):
        if self.notifying:
            logger.debug('Already notifying, nothing to do')
            return
        logger.debug('Starting notifications for token validation')
        self.notifying = True

    def StopNotify(self):
        if not self.notifying:
            logger.debug('Not notifying, nothing to do')
            return
        logger.debug('Stopping notifications for token validation')
        self.notifying = False

class CharacteristicUserDescriptionDescriptor(Descriptor):
    """
    Writable CUD descriptor.
    """

    CUD_UUID = "2901"

    def __init__(
        self, bus, index, characteristic,
    ):
        self.writable = 'writable-auxiliaries' in characteristic.flags
        self.value = array.array("B", characteristic.description)
        self.value = self.value.tolist()
        Descriptor.__init__(self, bus, index, self.CUD_UUID, ["read"], characteristic)

    def ReadValue(self, options):
        return self.value

    def WriteValue(self, value, options):
        if not self.writable:
            logger.error("Error User Description not writable")
        self.value = value

class TeddyAdvertisement(Advertisement):
    def __init__(self, bus, index):
        Advertisement.__init__(self, bus, index, "peripheral")
        self.add_manufacturer_data(
            0xFFFF, [0x00, 0x00], # Placeholders for now, we can use these for beaconing
        )
        self.add_service_uuid(TeddyS1Service.TEDDY_SVC_UUID)

        self.add_local_name(get_adv_name())
        self.include_tx_power = True

def register_ad_cb():
    logger.info("Advertisement registered")


def register_ad_error_cb(error):
    logger.critical("Failed to register advertisement: %s", str(error))
    mainloop.quit()

def get_adv_name():
    last_four = deviceManager.deviceId[-4:]
    adv_name = "Teddy-" + last_four
    return adv_name

def main():
    global mainloop

    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

    # get the system bus
    bus = dbus.SystemBus()
    # get the ble controller
    adapter = find_adapter(bus)

    if not adapter:
        logger.critical("GattManager1 interface not found")
        return

    adapter_obj = bus.get_object(BLUEZ_SERVICE_NAME, adapter)

    adapter_props = dbus.Interface(adapter_obj, "org.freedesktop.DBus.Properties")

    # powered property on the controller to on
    adapter_props.Set("org.bluez.Adapter1", "Powered", dbus.Boolean(1))

    # set bt alias to "Teddy-" plus the last 4 hex chars of the device ID
    adapter_props.Set('org.bluez.Adapter1', 'Alias', get_adv_name())

    # Get manager objs
    service_manager = dbus.Interface(adapter_obj, GATT_MANAGER_IFACE)
    ad_manager = dbus.Interface(adapter_obj, LE_ADVERTISING_MANAGER_IFACE)

    advertisement = TeddyAdvertisement(bus, 0)

    app = Application(bus)
    app.add_service(TeddyS1Service(bus, 2))

    mainloop = MainLoop()

    ad_manager.RegisterAdvertisement(
        advertisement.get_path(),
        {},
        reply_handler=register_ad_cb,
        error_handler=register_ad_error_cb,
    )

    logger.info("Registering GATT application...")

    service_manager.RegisterApplication(
        app.get_path(),
        {},
        reply_handler=register_app_cb,
        error_handler=[register_app_error_cb],
    )

    mainloop.run()

if __name__ == "__main__":
    main()
